@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2018 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2021 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -74,6 +74,9 @@ static bool resave_fs;
 static NODE *save_FS;		/* save current value of FS when line is read,
 				 * to be used in deferred parsing
 				 */
+static NODE *save_FPAT;		/* save current value of FPAT when line is read,
+				 * to be used in deferred parsing
+				 */
 static awk_fieldwidth_info_t *FIELDWIDTHS = NULL;
 
 NODE **fields_arr;		/* array of pointers to the field nodes */
@@ -87,6 +90,8 @@ Regexp *FPAT_re_no_case = NULL;
 Regexp *FPAT_regexp = NULL;
 NODE *Null_field = NULL;
 
+#define clear_mpfr(n) ((n)->flags &= ~(MPFN | MPZN | NUMCUR))
+
 /* init_fields --- set up the fields array to start with */
 
 void
@@ -94,16 +99,13 @@ init_fields()
 {
 	emalloc(fields_arr, NODE **, sizeof(NODE *), "init_fields");
 
-	getnode(fields_arr[0]);
-	*fields_arr[0] = *Nnull_string;
+	fields_arr[0] = make_string("", 0);
 	fields_arr[0]->flags |= NULL_FIELD;
 
 	parse_extent = fields_arr[0]->stptr;
 	save_FS = dupnode(FS_node->var_value);
 
-	getnode(Null_field);
-	*Null_field = *Nnull_string;
-	Null_field->valref = 1;
+	Null_field = make_string("", 0);
 	Null_field->flags = (STRCUR|STRING|NULL_FIELD); /* do not set MALLOC */
 
 	field0_valid = true;
@@ -230,6 +232,7 @@ rebuild_record()
 			}
 
 			n->stptr = cops;
+			clear_mpfr(n);
 			unref(r);
 			fields_arr[i] = n;
 			assert((n->flags & WSTRCUR) == 0);
@@ -288,7 +291,9 @@ set_record(const char *buf, int cnt, const awk_fieldwidth_info_t *fw)
 		memset(databuf, '\0', databuf_size);
 	}
 	/* copy the data */
-	memcpy(databuf, buf, cnt);
+	if (cnt != 0) {
+		memcpy(databuf, buf, cnt);
+	}
 
 	/*
 	 * Add terminating '\0' so that C library routines
@@ -345,6 +350,11 @@ reset_record()
 	}
 }
 
+/*
+ * purge_record --- throw away the fields, make sure that
+ * 	individual nodes remain valid.
+ */
+
 static void
 purge_record()
 {
@@ -464,7 +474,9 @@ re_parse_field(long up_to,	/* parse only up to this field number */
 	if (len == 0)
 		return nf;
 
-	if (RS_is_null && default_FS) {
+	bool default_field_splitting = (RS_is_null && default_FS);
+
+	if (default_field_splitting) {
 		sep = scan;
 		while (scan < end && (*scan == ' ' || *scan == '\t' || *scan == '\n'))
 			scan++;
@@ -505,7 +517,7 @@ re_parse_field(long up_to,	/* parse only up to this field number */
            			(long) (REEND(rp, scan) - RESTART(rp, scan)), sep_arr);
 		scan += REEND(rp, scan);
 		field = scan;
-		if (scan == end)	/* FS at end of record */
+		if (scan == end && ! default_field_splitting)	/* FS at end of record */
 			(*set)(++nf, field, 0L, n);
 	}
 	if (nf != up_to && scan < end) {
@@ -841,6 +853,8 @@ get_field(long requested, Func_ptr *assign)
 	bool in_middle = false;
 	static bool warned = false;
 	extern int currule;
+	NODE *saved_fs;
+	Regexp *fs_regexp;
 
 	if (do_lint && currule == END && ! warned) {
 		warned = true;
@@ -855,10 +869,18 @@ get_field(long requested, Func_ptr *assign)
 		if (! field0_valid) {
 			/* first, parse remainder of input record */
 			if (NF == -1) {
+				in_middle = (parse_high_water != 0);
+				if (current_field_sep() == Using_FPAT) {
+					saved_fs = save_FPAT;
+					fs_regexp = FPAT_regexp;
+				} else {
+					saved_fs = save_FS;
+					fs_regexp = FS_regexp;
+				}
 				NF = (*parse_field)(UNLIMITED - 1, &parse_extent,
 		    			fields_arr[0]->stlen -
 					(parse_extent - fields_arr[0]->stptr),
-		    			save_FS, FS_regexp, set_field,
+					saved_fs, fs_regexp, set_field,
 					(NODE *) NULL,
 					(NODE *) NULL,
 					in_middle);
@@ -915,17 +937,6 @@ get_field(long requested, Func_ptr *assign)
 		 */
 		if (parse_extent == fields_arr[0]->stptr + fields_arr[0]->stlen)
 			NF = parse_high_water;
-		else if (parse_field == fpat_parse_field) {
-			/* FPAT parsing is weird, isolate the special cases */
-			char *rec_start = fields_arr[0]->stptr;
-			char *rec_end = fields_arr[0]->stptr + fields_arr[0]->stlen;
-
-			if (    parse_extent > rec_end
-			    || (parse_extent > rec_start && parse_extent < rec_end && requested == UNLIMITED-1))
-				NF = parse_high_water;
-			else if (parse_extent == rec_start) /* could be no match for FPAT */
-				NF = 0;
-		}
 		if (requested == UNLIMITED - 1)	/* UNLIMITED-1 means set NF */
 			requested = parse_high_water;
 	}
@@ -977,7 +988,9 @@ do_split(int nargs)
 		sep_arr = POP_PARAM();
 		if (sep_arr->type != Node_var_array)
 			fatal(_("split: fourth argument is not an array"));
-		if ((do_lint || do_lint_old) && ! warned) {
+		check_symtab_functab(sep_arr, "split",
+				_("%s: cannot use %s as fourth argument"));
+		if ((do_lint_extensions || do_lint_old) && ! warned) {
 			warned = true;
 			lintwarn(_("split: fourth argument is a gawk extension"));
 		}
@@ -987,6 +1000,8 @@ do_split(int nargs)
 	arr = POP_PARAM();
 	if (arr->type != Node_var_array)
 		fatal(_("split: second argument is not an array"));
+	check_symtab_functab(arr, "split",
+			_("%s: cannot use %s as second argument"));
 
 	if (sep_arr != NULL) {
 		if (sep_arr == arr)
@@ -1070,11 +1085,15 @@ do_patsplit(int nargs)
 		sep_arr = POP_PARAM();
 		if (sep_arr->type != Node_var_array)
 			fatal(_("patsplit: fourth argument is not an array"));
+		check_symtab_functab(sep_arr, "patsplit",
+				_("%s: cannot use %s as fourth argument"));
 	}
 	sep = POP();
 	arr = POP_PARAM();
 	if (arr->type != Node_var_array)
 		fatal(_("patsplit: second argument is not an array"));
+	check_symtab_functab(arr, "patsplit",
+			_("%s: cannot use %s as second argument"));
 
 	src = TOP_STRING();
 
@@ -1143,7 +1162,7 @@ set_FIELDWIDTHS()
 	bool fatal_error = false;
 	NODE *tmp;
 
-	if (do_lint && ! warned) {
+	if (do_lint_extensions && ! warned) {
 		warned = true;
 		lintwarn(_("`FIELDWIDTHS' is a gawk extension"));
 	}
@@ -1306,13 +1325,13 @@ choose_fs_function:
 
 		set_parser(null_parse_field);
 
-		if (do_lint && ! warned) {
+		if (do_lint_extensions && ! warned) {
 			warned = true;
 			lintwarn(_("null string for `FS' is a gawk extension"));
 		}
-	} else if (fs->stlen > 1) {
+	} else if (fs->stlen > 1 || (fs->flags & REGEX) != 0) {
 		if (do_lint_old)
-			warning(_("old awk does not support regexps as value of `FS'"));
+			lintwarn(_("old awk does not support regexps as value of `FS'"));
 		set_parser(re_parse_field);
 	} else if (RS_is_null) {
 		/* we know that fs->stlen <= 1 */
@@ -1433,11 +1452,10 @@ void
 set_FPAT()
 {
 	static bool warned = false;
-	static NODE *save_fpat = NULL;
 	bool remake_re = true;
 	NODE *fpat;
 
-	if (do_lint && ! warned) {
+	if (do_lint_extensions && ! warned) {
 		warned = true;
 		lintwarn(_("`FPAT' is a gawk extension"));
 	}
@@ -1456,9 +1474,9 @@ set_FPAT()
 	 * This comparison can't use cmp_nodes(), which pays attention
 	 * to IGNORECASE, and that's not what we want.
 	 */
-	if (save_fpat
-		&& FPAT_node->var_value->stlen == save_fpat->stlen
-		&& memcmp(FPAT_node->var_value->stptr, save_fpat->stptr, save_fpat->stlen) == 0) {
+	if (save_FPAT
+		&& FPAT_node->var_value->stlen == save_FPAT->stlen
+		&& memcmp(FPAT_node->var_value->stptr, save_FPAT->stptr, save_FPAT->stlen) == 0) {
 		if (FPAT_regexp != NULL)
 			FPAT_regexp = (IGNORECASE ? FPAT_re_no_case : FPAT_re_yes_case);
 
@@ -1471,8 +1489,8 @@ set_FPAT()
 		}
 	}
 
-	unref(save_fpat);
-	save_fpat = dupnode(FPAT_node->var_value);
+	unref(save_FPAT);
+	save_FPAT = dupnode(FPAT_node->var_value);
 	refree(FPAT_re_yes_case);
 	refree(FPAT_re_no_case);
 	FPAT_re_yes_case = FPAT_re_no_case = FPAT_regexp = NULL;
@@ -1565,7 +1583,7 @@ incr_scan(char **scanp, size_t len, mbstate_t *mbs)
  *         field_found = match(substr(string, parse_start), pattern)
  *         
  *         # check for an invalid null field and retry one character away
- *         if (nf > 0 && field_found && RSTART==1 && RLENGTH==0) {
+ *         if (nf > 0 && field_found && RSTART == 1 && RLENGTH == 0) {
  *             parse_start++
  *             field_found = match(substr(string, parse_start), pattern)
  *         }
