@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 2012-2019 the Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019, 2021, the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -25,21 +25,13 @@
 
 #include "awk.h"
 
-#ifdef HAVE_MPFR
-#define getmpfr(n)	getblock(n, BLOCK_MPFR, mpfr_ptr)
-#define freempfr(n)	freeblock(n, BLOCK_MPFR)
-
-#define getmpz(n)	getblock(n, BLOCK_MPZ, mpz_ptr)
-#define freempz(n)	freeblock(n, BLOCK_MPZ)
-#endif
-
 /* Declare some globals used by api_get_file: */
 extern IOBUF *curfile;
 extern INSTRUCTION *main_beginfile;
 extern int currule;
 
 static awk_bool_t node_to_awk_value(NODE *node, awk_value_t *result, awk_valtype_t wanted);
-static char *valtype2str(awk_valtype_t type);
+static const char *valtype2str(awk_valtype_t type);
 static NODE *ns_lookup(const char *name_space, const char *name, char **full_name);
 
 /*
@@ -79,7 +71,7 @@ api_get_argument(awk_ext_id_t id, size_t count,
 	/* if type is undefined */
 	if (arg->type == Node_var_new) {
 		if (wanted == AWK_UNDEFINED)
-			return true;
+			return awk_true;
 		else if (wanted == AWK_ARRAY) {
 			goto array;
 		} else {
@@ -90,7 +82,7 @@ api_get_argument(awk_ext_id_t id, size_t count,
 	/* at this point, we have real type */
 	if (arg->type == Node_var_array || arg->type == Node_array_ref) {
 		if (wanted != AWK_ARRAY && wanted != AWK_UNDEFINED)
-			return false;
+			return awk_false;
 		goto array;
 	} else
 		goto scalar;
@@ -156,6 +148,7 @@ awk_value_to_node(const awk_value_t *retval)
 {
 	NODE *ext_ret_val = NULL;
 	NODE *v;
+	int tval = 0;
 
 	if (retval == NULL)
 		fatal(_("awk_value_to_node: received null retval"));
@@ -177,8 +170,8 @@ awk_value_to_node(const awk_value_t *retval)
 			if (! do_mpfr)
 				fatal(_("awk_value_to_node: not in MPFR mode"));
 			ext_ret_val = make_number_node(MPFN);
-			memcpy(&ext_ret_val->mpg_numbr, retval->num_ptr, sizeof(ext_ret_val->mpg_numbr));
-			freempfr(retval->num_ptr);
+			tval = mpfr_set(ext_ret_val->mpg_numbr, (mpfr_ptr) retval->num_ptr, ROUND_MODE);
+			IEEE_FMT(ext_ret_val->mpg_numbr, tval);
 #else
 			fatal(_("awk_value_to_node: MPFR not supported"));
 #endif
@@ -188,8 +181,7 @@ awk_value_to_node(const awk_value_t *retval)
 			if (! do_mpfr)
 				fatal(_("awk_value_to_node: not in MPFR mode"));
 			ext_ret_val = make_number_node(MPZN);
-			memcpy(&ext_ret_val->mpg_i, retval->num_ptr, sizeof(ext_ret_val->mpg_i));
-			freempz(retval->num_ptr);
+			mpz_set(ext_ret_val->mpg_i, (mpz_ptr) retval->num_ptr);
 #else
 			fatal(_("awk_value_to_node: MPFR not supported"));
 #endif
@@ -284,7 +276,7 @@ api_lintwarn(awk_ext_id_t id, const char *format, ...)
 	(void) id;
 
 	va_start(args, format);
-	if (lintwarn == r_fatal) {
+	if (lintfunc == r_fatal) {
 		err(true, _("fatal: "), format, args);
 	} else {
 		err(false, _("warning: "), format, args);
@@ -520,7 +512,7 @@ assign_number(NODE *node, awk_value_t *val)
 		val->num_ptr = &node->mpg_i;
 		break;
 	default:
-		fatal(_("node_to_awk_value: detected invalid numeric flags combination `%s'; please file a bug report."), flags2str(node->flags));
+		fatal(_("node_to_awk_value: detected invalid numeric flags combination `%s'; please file a bug report"), flags2str(node->flags));
 		break;
 	}
 #endif
@@ -607,7 +599,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -641,7 +633,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -668,7 +660,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -701,7 +693,7 @@ node_to_awk_value(NODE *node, awk_value_t *val, awk_valtype_t wanted)
 				}
 				/* fall through */
 			default:
-				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report."), flags2str(node->flags));
+				warning(_("node_to_awk_value detected invalid flags combination `%s'; please file a bug report"), flags2str(node->flags));
 				val->val_type = AWK_UNDEFINED;
 				break;
 			}
@@ -1237,7 +1229,7 @@ api_release_flattened_array(awk_ext_id_t id,
 		awk_array_t a_cookie,
 		awk_flat_array_t *data)
 {
-	NODE *array = a_cookie;
+	NODE *array = (NODE *) a_cookie;
 	NODE **list;
 	size_t i, j, k;
 
@@ -1311,7 +1303,7 @@ api_get_mpfr(awk_ext_id_t id)
 {
 #ifdef HAVE_MPFR
 	mpfr_ptr p;
-	getmpfr(p);
+	emalloc(p, mpfr_ptr, sizeof(mpfr_t), "api_get_mpfr");
 	mpfr_init(p);
 	return p;
 #else
@@ -1327,7 +1319,8 @@ api_get_mpz(awk_ext_id_t id)
 {
 #ifdef HAVE_MPFR
 	mpz_ptr p;
-	getmpz(p);
+	emalloc(p, mpz_ptr, sizeof (mpz_t), "api_get_mpz");
+
 	mpz_init(p);
 	return p;
 #else
@@ -1567,13 +1560,13 @@ print_ext_versions(void)
 
 /* valtype2str --- return a printable representation of a value type */
 
-static char *
+static const char *
 valtype2str(awk_valtype_t type)
 {
 	static char buf[100];
 
 	// Important: keep in same order as in gawkapi.h!
-	static char *values[] = {
+	static const char *values[] = {
 		"AWK_UNDEFINED",
 		"AWK_NUMBER",
 		"AWK_STRING",
