@@ -3,7 +3,8 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2019 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2021,
+ * the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -46,7 +47,6 @@ unfield(NODE **l, NODE **r)
 		(*l) = dupnode(*r);
 		DEREF(*r);
 	}
-	force_string(*l);
 }
 
 #define UNFIELD(l, r)	unfield(& (l), & (r))
@@ -100,9 +100,21 @@ top:
 		}
 #endif
 
-		switch ((op = pc->opcode)) {
+		op = pc->opcode;
+		if (do_itrace) {
+			fprintf(stderr, "+ %s\n", opcode2str(op));
+			fflush(stderr);
+		}
+
+		switch (op) {
 		case Op_rule:
-			currule = pc->in_rule;   /* for sole use in Op_K_next, Op_K_nextfile, Op_K_getline */
+			currule = pc->in_rule;   /* for use in Op_K_next, Op_K_nextfile, Op_K_getline */
+			// 8/2020: See node BEGINFILE/ENDFILE in the manual.  We clear the record
+			// since conceptually we are before reading a new record from the
+			// upcoming file but haven't read it yet.
+			if (currule == BEGINFILE)
+				set_record("", 0, NULL);
+
 			/* fall through */
 		case Op_func:
 			source = pc->source_file;
@@ -253,19 +265,30 @@ uninitialized_scalar:
 			t2 = mk_sub(pc->sub_count);
 			t1 = POP_ARRAY(false);
 
-			if (do_lint && in_array(t1, t2) == NULL) {
+			if (in_array(t1, t2) == NULL) {
 				t2 = force_string(t2);
-				lintwarn(_("reference to uninitialized element `%s[\"%.*s\"]'"),
-					array_vname(t1), (int) t2->stlen, t2->stptr);
-				if (t2->stlen == 0)
-					lintwarn(_("subscript of array `%s' is null string"), array_vname(t1));
+
+				if (t1 == func_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"FUNCTAB", (int) t2->stlen, t2->stptr);
+				} else if (t1 == symbol_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"SYMTAB", (int) t2->stlen, t2->stptr);
+				} else if (do_lint) {
+					lintwarn(_("reference to uninitialized element `%s[\"%.*s\"]'"),
+						array_vname(t1), (int) t2->stlen, t2->stptr);
+					if (t2->stlen == 0)
+						lintwarn(_("subscript of array `%s' is null string"), array_vname(t1));
+				}
 			}
 
+			// continue the regular processing
+
 			/* for FUNCTAB, get the name as the element value */
 			if (t1 == func_table) {
 				static bool warned = false;
 
-				if (do_lint && ! warned) {
+				if (do_lint_extensions && ! warned) {
 					warned = true;
 					lintwarn(_("FUNCTAB is a gawk extension"));
 				}
@@ -283,12 +306,17 @@ uninitialized_scalar:
 			if (t1 == symbol_table) {
 				static bool warned = false;
 
-				if (do_lint && ! warned) {
+				if (do_lint_extensions && ! warned) {
 					warned = true;
 					lintwarn(_("SYMTAB is a gawk extension"));
 				}
 				if (r->type == Node_var)
 					r = r->var_value;
+				else if (r->type == Node_var_new) {
+					// variable may exist but have never been set.
+					r->var_value = dupnode(Nnull_string);
+					r = r->var_value;
+				}
 			}
 
 			if (r->type == Node_val)
@@ -397,16 +425,21 @@ uninitialized_scalar:
 					lintwarn(_("assignment used in conditional context"));
 					break;
 
-				case LINT_no_effect:
-					lintwarn(_("statement has no effect"));
-					break;
-
 				default:
 					cant_happen();
 				}
 			}
 			break;
 
+		case Op_lint_plus:
+			// no need to check do_lint, this opcode won't
+			// be generated if that's not true
+			t1 = TOP();
+			t2 = PEEK(1);
+			if ((t1->flags & STRING) != 0 && (t2->flags & STRING) != 0)
+				lintwarn(_("operator `+' used on two string values"));
+			break;
+
 		case Op_K_break:
 		case Op_K_continue:
 		case Op_jmp:
@@ -729,6 +762,8 @@ mod:
 			unref(*lhs);
 			r = POP_SCALAR();
 			UNFIELD(*lhs, r);
+			/* field variables need the string representation: */
+			force_string(*lhs);
 		}
 			break;
 
@@ -740,7 +775,10 @@ mod:
 
 			if (t1 != *lhs) {
 				unref(*lhs);
-				*lhs = dupnode(t1);
+				if (t1->valref == 1)
+					*lhs = t1;
+				else
+					*lhs = dupnode(t1);
 			}
 
 			if (t1 != t2 && t1->valref == 1 && (t1->flags & (MALLOC|MPFN|MPZN)) == MALLOC) {
@@ -1012,7 +1050,8 @@ arrayfor:
 						(unsigned long) max_expect);
 
 			PUSH_CODE(pc);
-			r = awk_value_to_node(pc->extfunc(arg_count, & result, f));
+			awk_value_t *ef_ret = pc->extfunc(arg_count, & result, f);
+			r = awk_value_to_node(ef_ret);
 			(void) POP_CODE();
 			while (arg_count-- > 0) {
 				t1 = POP();
@@ -1064,6 +1103,10 @@ match_re:
 			if (op != Op_match_rec) {
 				decr_sp();
 				DEREF(t1);
+				if (m->type == Node_dynregex) {
+				 	DEREF(m->re_exp);
+					m->re_exp = NULL;
+				}
 			}
 			r = node_Boolean[di];
 			UPREF(r);
@@ -1214,7 +1257,7 @@ match_re:
 			JUMPTO(ni);
 
 		case Op_K_getline_redir:
-			r = do_getline_redir(pc->into_var, pc->redir_type);
+			r = do_getline_redir(pc->into_var, (enum redirval) pc->redir_type);
 			PUSH(r);
 			break;
 
